@@ -1,61 +1,132 @@
 {-# LANGUAGE BangPatterns        #-}
 {-# LANGUAGE CPP                 #-}
+{-# LANGUAGE DeriveDataTypeable  #-}
 {-# LANGUAGE OverloadedStrings   #-}
 {-# LANGUAGE ScopedTypeVariables #-}
-
-{-|
-
-This module exports the 'Config' datatype, which you can use to configure the
-Snap HTTP server.
-
--}
-
-module Snap.Internal.Http.Server.Config where
-
 ------------------------------------------------------------------------------
-import           Blaze.ByteString.Builder
-import           Blaze.ByteString.Builder.Char8
-import           Control.Exception (SomeException)
-import           Control.Monad
-import qualified Data.ByteString.Char8 as B
-import           Data.ByteString (ByteString)
-import           Data.Char
-import           Data.Function
-import           Data.List
-import           Data.Maybe
-import           Data.Monoid
-import qualified Data.Text as T
-import qualified Data.Text.Encoding as T
-import           Data.Typeable
-import           Network(Socket)
-#if !MIN_VERSION_base(4,6,0)
-import           Prelude hiding (catch)
-#endif
-import           Snap.Core
-import           Snap.Iteratee ((>==>), enumBuilder)
-import           Snap.Internal.Debug (debug)
-import           Snap.Util.Proxy
-import           System.Console.GetOpt
-import           System.Environment hiding (getEnv)
-#ifndef PORTABLE
-import           System.Posix.Env
-#endif
-import           System.Exit
-import           System.IO
-------------------------------------------------------------------------------
-import           Snap.Internal.Http.Server (requestErrorMessage)
-
-
-------------------------------------------------------------------------------
--- | This datatype allows you to override which backend (either simple or
--- libev) to use. Most users will not want to set this, preferring to rely on
--- the compile-type default.
+-- | This module exports the 'Config' datatype, which you can use to configure
+-- the Snap HTTP server.
 --
--- Note that if you specify the libev backend and have not compiled in support
--- for it, your server will fail at runtime.
-data ConfigBackend = ConfigSimpleBackend
-                   | ConfigLibEvBackend
-  deriving (Show, Eq)
+module Snap.Internal.Http.Server.Config
+  -- NOTE: also edit Snap.Http.Server.Config if you change these
+  ( ConfigLog(..)
+  , Config(..)
+  , ProxyType(..)
+
+  , emptyConfig
+  , defaultConfig
+
+  , commandLineConfig
+  , extendedCommandLineConfig
+  , completeConfig
+
+  , optDescrs
+  , fmapOpt
+
+  , getAccessLog
+  , getBind
+  , getCompression
+  , getDefaultTimeout
+  , getErrorHandler
+  , getErrorLog
+  , getHostname
+  , getLocale
+  , getOther
+  , getPort
+  , getProxyType
+  , getSSLBind
+  , getSSLCert
+  , getSSLChainCert
+  , getSSLKey
+  , getSSLPort
+  , getVerbose
+  , getStartupHook
+  , getUnixSocket
+  , getUnixSocketAccessMode
+
+  , setAccessLog
+  , setBind
+  , setCompression
+  , setDefaultTimeout
+  , setErrorHandler
+  , setErrorLog
+  , setHostname
+  , setLocale
+  , setOther
+  , setPort
+  , setProxyType
+  , setSSLBind
+  , setSSLCert
+  , setSSLChainCert
+  , setSSLKey
+  , setSSLPort
+  , setVerbose
+  , setUnixSocket
+  , setUnixSocketAccessMode
+
+  , setStartupHook
+
+  , StartupInfo(..)
+  , getStartupSockets
+  , getStartupConfig
+
+  -- * Private
+  , emptyStartupInfo
+  , setStartupSockets
+  , setStartupConfig
+  ) where
+
+------------------------------------------------------------------------------
+import           Control.Exception          (SomeException)
+import           Control.Monad              (when)
+import           Data.Bits                  ((.&.))
+import           Data.ByteString            (ByteString)
+import qualified Data.ByteString.Char8      as S
+import qualified Data.ByteString.Lazy.Char8 as L
+import qualified Data.CaseInsensitive       as CI
+import           Data.Function              (on)
+import           Data.List                  (foldl')
+import           Data.Maybe                 (isJust, isNothing)
+#if !MIN_VERSION_base(4,8,0)
+import           Data.Monoid                (Monoid(..))
+#endif
+import           Data.Monoid                (Last (Last, getLast))
+import qualified Data.Text                  as T
+import qualified Data.Text.Encoding         as T
+#if MIN_VERSION_base(4,7,0)
+import           Data.Typeable              (Typeable)
+#else
+import           Data.Typeable              (TyCon, Typeable, Typeable1 (..), mkTyConApp, mkTyCon3)
+#endif
+import           Network                    (Socket)
+import           Numeric                    (readOct, showOct)
+#if !MIN_VERSION_base(4,6,0)
+import           Prelude                    hiding (catch)
+#endif
+import           System.Console.GetOpt      (ArgDescr (..), ArgOrder (Permute), OptDescr (..), getOpt, usageInfo)
+import           System.Environment         hiding (getEnv)
+#ifndef PORTABLE
+import           Data.Char                  (isAlpha)
+import           System.Posix.Env           (getEnv)
+#endif
+import           System.Exit                (exitFailure)
+import           System.IO                  (hPutStrLn, stderr)
+------------------------------------------------------------------------------
+import           Data.ByteString.Builder    (Builder, byteString, stringUtf8, toLazyByteString)
+import qualified System.IO.Streams          as Streams
+------------------------------------------------------------------------------
+import           Snap.Core                  (MonadSnap, Request (rqClientAddr, rqClientPort), emptyResponse, finishWith, getRequest, logError, setContentLength, setContentType, setResponseBody, setResponseStatus)
+import           Snap.Internal.Debug        (debug)
+
+
+------------------------------------------------------------------------------
+-- | FIXME
+--
+-- Note: this type changed in snap-server 1.0.0.0.
+data ProxyType = NoProxy
+               | HaProxy
+               | X_Forwarded_For
+  deriving (Show, Eq, Typeable)
 
 ------------------------------------------------------------------------------
 -- | Data type representing the configuration of a logging target
@@ -68,6 +139,54 @@ instance Show ConfigLog where
     show (ConfigFileLog f) = "log to file " ++ show f
     show (ConfigIoLog _)   = "custom logging handler"
 
+
+------------------------------------------------------------------------------
+-- We should be using ServerConfig here. There needs to be a clearer
+-- separation between:
+--
+--   * what the underlying code needs to configure itself
+--
+--   * what the command-line processing does.
+--
+-- The latter will provide "library" helper functions that operate on
+-- ServerConfig/etc in order to allow users to configure their own environment.
+--
+--
+-- Todo:
+--
+--  * need a function ::
+--      CommandLineConfig -> IO [(ServerConfig hookState, AcceptFunc)]
+--
+--       this will prep for another function that will spawn all of the
+--       accept loops with httpAcceptLoop.
+--
+--  * all backends provide "Some -> Foo -> Config -> IO AcceptFunc"
+--
+--  * add support for socket activation to command line, or delegate to
+--    different library? It's linux-only anyways, need to ifdef. It would be
+--    silly to depend on the socket-activation library for that one little
+--    function.
+--
+--  * break config into multiple modules:
+--
+--     * everything that modifies the snap handler (compression, proxy
+--       settings, error handler)
+--
+--     * everything that directly modifies server settings (hostname /
+--       defaultTimeout / hooks / etc)
+--
+--     * everything that configures backends (port/bind/ssl*)
+--
+--     * everything that handles command line stuff
+--
+--     * utility stuff
+--
+-- Cruft that definitely must be removed:
+--
+--  * ConfigLog -- this becomes a binary option on the command-line side (no
+--    logging or yes, to this file), but the ConfigIoLog gets zapped
+--    altogether.
+
 ------------------------------------------------------------------------------
 -- | A record type which represents partial configurations (for 'httpServe')
 -- by wrapping all of its fields in a 'Maybe'. Values of this type are usually
@@ -89,12 +208,13 @@ data Config m a = Config
     , sslcert        :: Maybe FilePath
     , sslchaincert   :: Maybe Bool
     , sslkey         :: Maybe FilePath
+    , unixsocket     :: Maybe FilePath
+    , unixaccessmode :: Maybe Int
     , compression    :: Maybe Bool
     , verbose        :: Maybe Bool
     , errorHandler   :: Maybe (SomeException -> m ())
     , defaultTimeout :: Maybe Int
     , other          :: Maybe a
-    , backend        :: Maybe ConfigBackend
     , proxyType      :: Maybe ProxyType
     , startupHook    :: Maybe (StartupInfo m a -> IO ())
     }
@@ -102,12 +222,11 @@ data Config m a = Config
   deriving Typeable
 #else
 
-
 ------------------------------------------------------------------------------
 -- | The 'Typeable1' instance is here so 'Config' values can be
 -- dynamically loaded with Hint.
 configTyCon :: TyCon
-configTyCon = mkTyCon "Snap.Http.Server.Config.Config"
+configTyCon = mkTyCon3 "snap-server" "Snap.Http.Server.Config" "Config"
 {-# NOINLINE configTyCon #-}
 
 instance (Typeable1 m) => Typeable1 (Config m) where
@@ -128,10 +247,11 @@ instance Show (Config m a) where
                      , "sslcert: "        ++ _sslcert
                      , "sslchaincert: "   ++ _sslchaincert
                      , "sslkey: "         ++ _sslkey
+                     , "unixsocket: "     ++ _unixsocket
+                     , "unixaccessmode: " ++ _unixaccessmode
                      , "compression: "    ++ _compression
                      , "verbose: "        ++ _verbose
                      , "defaultTimeout: " ++ _defaultTimeout
-                     , "backend: "        ++ _backend
                      , "proxyType: "      ++ _proxyType
                      ]
 
@@ -150,8 +270,11 @@ instance Show (Config m a) where
         _compression    = show $ compression    c
         _verbose        = show $ verbose        c
         _defaultTimeout = show $ defaultTimeout c
-        _backend        = show $ backend        c
         _proxyType      = show $ proxyType      c
+        _unixsocket     = show $ unixsocket     c
+        _unixaccessmode = case unixaccessmode c of
+                               Nothing -> "Nothing"
+                               Just s -> ("Just 0" ++) . showOct s $ []
 
 
 ------------------------------------------------------------------------------
@@ -175,12 +298,13 @@ instance Monoid (Config m a) where
         , sslcert        = Nothing
         , sslchaincert   = Nothing
         , sslkey         = Nothing
+        , unixsocket     = Nothing
+        , unixaccessmode = Nothing
         , compression    = Nothing
         , verbose        = Nothing
         , errorHandler   = Nothing
         , defaultTimeout = Nothing
         , other          = Nothing
-        , backend        = Nothing
         , proxyType      = Nothing
         , startupHook    = Nothing
         }
@@ -197,16 +321,18 @@ instance Monoid (Config m a) where
         , sslcert        = ov sslcert
         , sslchaincert   = ov sslchaincert
         , sslkey         = ov sslkey
+        , unixsocket     = ov unixsocket
+        , unixaccessmode = ov unixaccessmode
         , compression    = ov compression
         , verbose        = ov verbose
         , errorHandler   = ov errorHandler
         , defaultTimeout = ov defaultTimeout
         , other          = ov other
-        , backend        = ov backend
         , proxyType      = ov proxyType
         , startupHook    = ov startupHook
         }
       where
+        ov :: (Config m a -> Maybe b) -> Maybe b
         ov f = getLast $! (mappend `on` (Last . f)) a b
 
 
@@ -222,10 +348,10 @@ defaultConfig = mempty
     , verbose        = Just True
     , errorHandler   = Just defaultErrorHandler
     , bind           = Just "0.0.0.0"
-    , sslbind        = Just "0.0.0.0"
-    , sslcert        = Just "cert.pem"
-    , sslchaincert   = Just False
-    , sslkey         = Just "key.pem"
+    , sslbind        = Nothing
+    , sslcert        = Nothing
+    , sslkey         = Nothing
+    , sslchaincert   = Nothing
     , defaultTimeout = Just 60
     }
 
@@ -280,6 +406,24 @@ getSSLChainCert = sslchaincert
 getSSLKey         :: Config m a -> Maybe FilePath
 getSSLKey = sslkey
 
+-- | File path to unix socket. Must be absolute path, but allows for symbolic
+-- links.
+getUnixSocket     :: Config m a -> Maybe FilePath
+getUnixSocket = unixsocket
+
+-- | Access mode for unix socket, by default is system specific.
+-- This should only be used to grant additional permissions to created
+-- socket file, and not to remove permissions set by default.
+-- The only portable way to limit access to socket is creating it in a
+-- directory with proper permissions set.
+--
+-- Most BSD systems ignore access permissions on unix sockets.
+--
+-- Note: This uses umask. There is a race condition if process creates other
+-- files at the same time as opening a unix socket with this option set.
+getUnixSocketAccessMode :: Config m a -> Maybe Int
+getUnixSocketAccessMode = unixaccessmode
+
 -- | If set and set to True, compression is turned on when applicable
 getCompression    :: Config m a -> Maybe Bool
 getCompression = compression
@@ -298,9 +442,6 @@ getDefaultTimeout = defaultTimeout
 getOther :: Config m a -> Maybe a
 getOther = other
 
-getBackend :: Config m a -> Maybe ConfigBackend
-getBackend = backend
-
 getProxyType :: Config m a -> Maybe ProxyType
 getProxyType = proxyType
 
@@ -346,6 +487,12 @@ setSSLChainCert x c = c { sslchaincert = Just x }
 setSSLKey         :: FilePath                -> Config m a -> Config m a
 setSSLKey x c = c { sslkey = Just x }
 
+setUnixSocket     :: FilePath                -> Config m a -> Config m a
+setUnixSocket x c = c { unixsocket = Just x }
+
+setUnixSocketAccessMode :: Int               -> Config m a -> Config m a
+setUnixSocketAccessMode p c = c { unixaccessmode = Just ( p .&. 0o777) }
+
 setCompression    :: Bool                    -> Config m a -> Config m a
 setCompression x c = c { compression = Just x }
 
@@ -361,9 +508,6 @@ setDefaultTimeout x c = c { defaultTimeout = Just x }
 setOther          :: a                       -> Config m a -> Config m a
 setOther x c = c { other = Just x }
 
-setBackend        :: ConfigBackend           -> Config m a -> Config m a
-setBackend x c = c { backend = Just x }
-
 setProxyType      :: ProxyType               -> Config m a -> Config m a
 setProxyType x c = c { proxyType = Just x }
 
@@ -375,14 +519,15 @@ setStartupHook x c = c { startupHook = Just x }
 
 -- | Arguments passed to 'setStartupHook'.
 data StartupInfo m a = StartupInfo
-    { startupHookConfig :: Config m a
+    { startupHookConfig  :: Config m a
     , startupHookSockets :: [Socket]
     }
 
 emptyStartupInfo :: StartupInfo m a
 emptyStartupInfo = StartupInfo emptyConfig []
 
--- | The the 'Socket's opened by the server. There will be two 'Socket's for SSL connections, and one otherwise.
+-- | The 'Socket's opened by the server. There will be two 'Socket's for SSL
+-- connections, and one otherwise.
 getStartupSockets :: StartupInfo m a -> [Socket]
 getStartupSockets = startupHookSockets
 
@@ -432,31 +577,31 @@ toString = T.unpack . T.decodeUtf8
 ------------------------------------------------------------------------------
 -- | Returns a description of the snap command line options suitable for use
 -- with "System.Console.GetOpt".
-optDescrs :: MonadSnap m =>
+optDescrs :: forall m a . MonadSnap m =>
              Config m a         -- ^ the configuration defaults.
           -> [OptDescr (Maybe (Config m a))]
 optDescrs defaults =
-    [ Option [] ["hostname"]
+    [ Option "" ["hostname"]
              (ReqArg (Just . setConfig setHostname . bsFromString) "NAME")
              $ "local hostname" ++ defaultC getHostname
-    , Option ['b'] ["address"]
+    , Option "b" ["address"]
              (ReqArg (\s -> Just $ mempty { bind = Just $ bsFromString s })
                      "ADDRESS")
              $ "address to bind to" ++ defaultO bind
-    , Option ['p'] ["port"]
+    , Option "p" ["port"]
              (ReqArg (\s -> Just $ mempty { port = Just $ read s}) "PORT")
              $ "port to listen on" ++ defaultO port
-    , Option [] ["ssl-address"]
+    , Option "" ["ssl-address"]
              (ReqArg (\s -> Just $ mempty { sslbind = Just $ bsFromString s })
                      "ADDRESS")
              $ "ssl address to bind to" ++ defaultO sslbind
-    , Option [] ["ssl-port"]
+    , Option "" ["ssl-port"]
              (ReqArg (\s -> Just $ mempty { sslport = Just $ read s}) "PORT")
              $ "ssl port to listen on" ++ defaultO sslport
-    , Option [] ["ssl-cert"]
+    , Option "" ["ssl-cert"]
              (ReqArg (\s -> Just $ mempty { sslcert = Just s}) "PATH")
              $ "path to ssl certificate in PEM format" ++ defaultO sslcert
-    , Option [] ["ssl-chain-cert"]
+   , Option [] ["ssl-chain-cert"]
              (NoArg $ Just $ setConfig setSSLChainCert True)
              $ "certificate file contains complete certificate chain" ++ defaultB sslchaincert "site certificate only" "complete certificate chain"
     , Option [] ["no-ssl-chain-cert"]
@@ -465,58 +610,88 @@ optDescrs defaults =
     , Option [] ["ssl-key"]
              (ReqArg (\s -> Just $ mempty { sslkey = Just s}) "PATH")
              $ "path to ssl private key in PEM format" ++ defaultO sslkey
-    , Option [] ["access-log"]
+    , Option "" ["access-log"]
              (ReqArg (Just . setConfig setAccessLog . ConfigFileLog) "PATH")
-             $ "access log" ++ (defaultC $ getAccessLog)
-    , Option [] ["error-log"]
+             $ "access log" ++ defaultC getAccessLog
+    , Option "" ["error-log"]
              (ReqArg (Just . setConfig setErrorLog . ConfigFileLog) "PATH")
-             $ "error log" ++ (defaultC $ getErrorLog)
-    , Option [] ["no-access-log"]
+             $ "error log" ++ defaultC getErrorLog
+    , Option "" ["no-access-log"]
              (NoArg $ Just $ setConfig setAccessLog ConfigNoLog)
-             $ "don't have an access log"
-    , Option [] ["no-error-log"]
+             "don't have an access log"
+    , Option "" ["no-error-log"]
              (NoArg $ Just $ setConfig setErrorLog ConfigNoLog)
-             $ "don't have an error log"
-    , Option ['c'] ["compression"]
+             "don't have an error log"
+    , Option "c" ["compression"]
              (NoArg $ Just $ setConfig setCompression True)
              $ "use gzip compression on responses" ++
                defaultB getCompression "compressed" "uncompressed"
-    , Option ['t'] ["timeout"]
+    , Option "t" ["timeout"]
              (ReqArg (\t -> Just $ mempty {
                               defaultTimeout = Just $ read t
                             }) "SECS")
              $ "set default timeout in seconds" ++ defaultC defaultTimeout
-    , Option [] ["no-compression"]
+    , Option "" ["no-compression"]
              (NoArg $ Just $ setConfig setCompression False)
              $ "serve responses uncompressed" ++
                defaultB compression "compressed" "uncompressed"
-    , Option ['v'] ["verbose"]
+    , Option "v" ["verbose"]
              (NoArg $ Just $ setConfig setVerbose True)
              $ "print server status updates to stderr" ++
                defaultC getVerbose
-    , Option ['q'] ["quiet"]
+    , Option "q" ["quiet"]
              (NoArg $ Just $ setConfig setVerbose False)
              $ "do not print anything to stderr" ++
                defaultB getVerbose "verbose" "quiet"
-    , Option [] ["proxy"]
-             (ReqArg (\t -> Just $ setConfig setProxyType $ read t)
+    , Option "" ["proxy"]
+             (ReqArg (Just . setConfig setProxyType . parseProxy . CI.mk)
                      "X_Forwarded_For")
-             $ concat [ "Set --proxy=X_Forwarded_For if your snap application "
-                      , "is behind an HTTP reverse proxy to ensure that "
-                      , "rqRemoteAddr is set properly."
+             $ concat [ "Set --proxy=X_Forwarded_For if your snap application \n"
+                      , "is behind an HTTP reverse proxy to ensure that \n"
+                      , "rqClientAddr is set properly.\n"
+                      , "Set --proxy=haproxy to use the haproxy protocol\n("
+                      , "http://haproxy.1wt.eu/download/1.5/doc/proxy-protocol.txt)"
                       , defaultC getProxyType ]
-    , Option ['h'] ["help"]
+    , Option "" ["unix-socket"]
+             (ReqArg (Just . setConfig setUnixSocket) "PATH")
+             $ concat ["Absolute path to unix socket file. "
+                      , "File will be removed if already exists"]
+    , Option "" ["unix-socket-mode"]
+             (ReqArg (Just . setConfig setUnixSocketAccessMode . parseOctal)
+                     "MODE")
+             $ concat ["Access mode for unix socket in octal, for example 0760.\n"
+                      ," Default is system specific."]
+    , Option "h" ["help"]
              (NoArg Nothing)
-             $ "display this help and exit"
+             "display this help and exit"
     ]
 
   where
+    parseProxy s | s == "NoProxy"         = NoProxy
+                 | s == "X_Forwarded_For" = X_Forwarded_For
+                 | s == "haproxy"         = HaProxy
+                 | otherwise = error $ concat [
+                         "Error (--proxy): expected one of 'NoProxy', "
+                       , "'X_Forwarded_For', or 'haproxy'. Got '"
+                       , CI.original s
+                       , "'"
+                       ]
+    parseOctal s = case readOct s of
+          ((v, _):_) | v >= 0 && v <= 0o777 -> v
+          _ -> error $ "Error (--unix-socket-mode): expected octal access mode"
+
     setConfig f c  = f c mempty
     conf           = defaultConfig `mappend` defaults
-    defaultB f y n = maybe "" (\b -> ", default " ++ if b
-                                                       then y
-                                                       else n) $ f conf :: String
+
+    defaultB :: (Config m a -> Maybe Bool) -> String -> String -> String
+    defaultB f y n = (maybe "" (\b -> ", default " ++ if b
+                                                        then y
+                                                        else n) $ f conf) :: String
+
+    defaultC :: (Show b) => (Config m a -> Maybe b) -> String
     defaultC f     = maybe "" ((", default " ++) . show) $ f conf
+
+    defaultO :: (Show b) => (Config m a -> Maybe b) -> String
     defaultO f     = maybe ", default off" ((", default " ++) . show) $ f conf
 
 
@@ -530,23 +705,24 @@ defaultErrorHandler e = do
     logError sm
 
     finishWith $ setContentType "text/plain; charset=utf-8"
-               . setContentLength (fromIntegral $ B.length msg)
+               . setContentLength (fromIntegral $ S.length msg)
                . setResponseStatus 500 "Internal Server Error"
-               . modifyResponseBody
-                     (>==> enumBuilder (fromByteString msg))
+               . setResponseBody errBody
                $ emptyResponse
 
   where
+    errBody os = Streams.write (Just msgB) os >> return os
+
+    toByteString = S.concat . L.toChunks . toLazyByteString
     smsg req = toByteString $ requestErrorMessage req e
 
     msg  = toByteString msgB
     msgB = mconcat [
-             fromByteString "A web handler threw an exception. Details:\n"
-           , fromShow e
+             byteString "A web handler threw an exception. Details:\n"
+           , stringUtf8 $ show e
            ]
 
 
-
 ------------------------------------------------------------------------------
 -- | Returns a 'Config' obtained from parsing command-line options, using the
 -- default Snap 'OptDescr' set.
@@ -640,3 +816,24 @@ fmapOpt :: (a -> b) -> OptDescr a -> OptDescr b
 fmapOpt f (Option s l d e) = Option s l (fmapArg f d) e
 
 
+------------------------------------------------------------------------------
+requestErrorMessage :: Request -> SomeException -> Builder
+requestErrorMessage req e =
+    mconcat [ byteString "During processing of request from "
+            , byteString $ rqClientAddr req
+            , byteString ":"
+            , fromShow $ rqClientPort req
+            , byteString "\nrequest:\n"
+            , fromShow $ show req
+            , byteString "\n"
+            , msgB
+            ]
+  where
+    msgB = mconcat [
+             byteString "A web handler threw an exception. Details:\n"
+           , fromShow e
+           ]
+
+------------------------------------------------------------------------------
+fromShow :: Show a => a -> Builder
+fromShow = stringUtf8 . show
